﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Threading;

namespace TestLibrary
{
    /// <summary>
    /// 비동기 가상화 컬렉션
    /// </summary>
    /// <typeparam name="TItem">항목 타입</typeparam>
    public class AsyncVirtualizingCollection<TItem> : VirtualizingCollection<TItem>, INotifyCollectionChanged, INotifyPropertyChanged where TItem : class
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 컬렉션 변경시 - CollectionChanged

        /// <summary>
        /// 컬렉션 변경시
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion
        #region 속성 변경시 - PropertyChanged

        /// <summary>
        /// 속성 변경시
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 동기화 컨텍스트
        /// </summary>
        private readonly SynchronizationContext synchronizationContext;

        /// <summary>
        /// 로딩 여부
        /// </summary>
        private bool isLoading;

        /// <summary>
        /// 초기화 여부
        /// </summary>
        private bool isInitializing;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 동기화 컨텍스트 - SynchronizationContext

        /// <summary>
        /// 동기화 컨텍스트
        /// </summary>
        protected SynchronizationContext SynchronizationContext
        {
            get
            {
                return this.synchronizationContext;
            }
        }

        #endregion
        #region 로딩 여부 - IsLoading

        /// <summary>
        /// 로딩 여부
        /// </summary>
        public bool IsLoading
        {
            get
            {
                return this.isLoading;
            }
            set
            {
                if(value != this.isLoading)
                {
                    this.isLoading = value;

                    FirePropertyChangedEvent("IsLoading");
                }
            }
        }

        #endregion
        #region 초기화 여부 - IsInitializing

        /// <summary>
        /// 초기화 여부
        /// </summary>
        public bool IsInitializing
        {
            get
            {
                return this.isInitializing;
            }
            set
            {
                if(value != this.isInitializing)
                {
                    this.isInitializing = value;

                    FirePropertyChangedEvent("IsInitializing");
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - AsyncVirtualizingCollection(itemProvider, pageSize, pageTimeout)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        /// <param name="pageSize">페이지 크기</param>
        /// <param name="pageTimeout">페이지 타임아웃</param>
        public AsyncVirtualizingCollection(IItemProvider<TItem> itemProvider, int pageSize, int pageTimeout) : base(itemProvider, pageSize, pageTimeout)
        {
            this.synchronizationContext = SynchronizationContext.Current;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 컬렉션 변경시 이벤트 발생시키기 - FireCollectionChangedEvent(e)

        /// <summary>
        /// 컬렉션 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected virtual void FireCollectionChangedEvent(NotifyCollectionChangedEventArgs e)
        {
            CollectionChanged?.Invoke(this, e);
        }

        #endregion
        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(e)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected virtual void FirePropertyChangedEvent(PropertyChangedEventArgs e)
        {
            PropertyChanged?.Invoke(this, e);
        }

        #endregion

        #region 카운트 로드하기 - LoadCount()

        /// <summary>
        /// 카운트 로드하기
        /// </summary>
        protected override void LoadCount()
        {
            if(Count == 0)
            {
                IsInitializing = true;
            }

            ThreadPool.QueueUserWorkItem(LoadCount);
        }

        #endregion
        #region 카운트 로드 완료시 처리하기 - ProcessLoadCountCompleted(argument)

        /// <summary>
        /// 카운트 로드 완료시 처리하기
        /// </summary>
        /// <param name="argument">인자</param>
        protected virtual void ProcessLoadCountCompleted(object argument)
        {
            int newCount = (int)argument;

            SetNewCount(newCount);

            IsInitializing = false;
        }

        #endregion
        #region 페이지 로드하기 - LoadPage(pageIndex, pageLength)

        /// <summary>
        /// 페이지 로드하기
        /// </summary>
        /// <param name="pageIndex">페이지 인덱스</param>
        /// <param name="pageLength">페이지 길이</param>
        protected override void LoadPage(int pageIndex, int pageLength)
        {
            IsLoading = true;

            ThreadPool.QueueUserWorkItem(LoadPage, new int[] { pageIndex, pageLength });
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 컬렉션 리셋하기 - ResetCollectionReset()

        /// <summary>
        /// 컬렉션 리셋하기
        /// </summary>
        private void ResetCollectionReset()
        {
            NotifyCollectionChangedEventArgs e = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset);

            FireCollectionChangedEvent(e);
        }

        #endregion
        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(propertyName)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="propertyName">속성명</param>
        private void FirePropertyChangedEvent(string propertyName)
        {
            PropertyChangedEventArgs e = new PropertyChangedEventArgs(propertyName);

            FirePropertyChangedEvent(e);
        }

        #endregion

        #region 카운트 로드하기 - LoadCount(argument)

        /// <summary>
        /// 카운트 로드하기
        /// </summary>
        /// <param name="argument">인자</param>
        private void LoadCount(object argument)
        {
            int count = FetchCount();

            SynchronizationContext.Send(ProcessLoadCountCompleted, count);
        }

        #endregion
        #region 신규 카운트 설정하기 - SetNewCount(newCount)

        /// <summary>
        /// 신규 카운트 설정하기
        /// </summary>
        /// <param name="newCount">신규 카운트</param>
        private void SetNewCount(int newCount)
        {
            if(newCount != Count)
            {
                Count = newCount;

                EmptyCache();

                ResetCollectionReset();
            }
        }

        #endregion
        #region 페이지 로드 완료시 처리하기 - ProcessLoadPageCompleted(state)

        /// <summary>
        /// 페이지 로드 완료시 처리하기
        /// </summary>
        /// <param name="state">상태</param>
        private void ProcessLoadPageCompleted(object state)
        {
            object[] argumentArray = (object[])state;

            int pageIndex = (int)argumentArray[0];

            IList<TItem> itemList = (IList<TItem>)argumentArray[1];

            int newCount = (int)argumentArray[2];

            SetNewCount(newCount);

            PopulatePage(pageIndex, itemList);

            IsLoading = false;
        }

        #endregion
        #region 페이지 로드하기 - LoadPage(state)

        /// <summary>
        /// 페이지 로드하기
        /// </summary>
        /// <param name="state">상태</param>
        private void LoadPage(object state)
        {
            int[] argumentArray = (int[])state;

            int pageIndex    = argumentArray[0];
            int pageLength   = argumentArray[1];
            int overallCount = 0;

            IList<TItem> itemList = FetchPage(pageIndex, pageLength, out overallCount);

            SynchronizationContext.Send(ProcessLoadPageCompleted, new object[] { pageIndex, itemList, overallCount });
        }

        #endregion
    }
}