﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Collections;

namespace TestLibrary
{
    /// <summary>
    /// 가상화 컬렉션
    /// </summary>
    /// <typeparam name="TItem">항목 타입</typeparam>
    public class VirtualizingCollection<TItem> : IList<DataWrapper<TItem>>, IList where TItem : class
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 항목 공급자
        /// </summary>
        private readonly IItemProvider<TItem> itemProvider;

        /// <summary>
        /// 페이지 크기
        /// </summary>
        private readonly int pageSize = 100;

        /// <summary>
        /// 페이지 타임아웃
        /// </summary>
        private readonly long pageTimeout = 10000;

        /// <summary>
        /// 카운트
        /// </summary>
        private int count = -1;

        /// <summary>
        /// 데이터 페이지 딕셔너리
        /// </summary>
        private Dictionary<int, DataPage<TItem>> pageDictionary = new Dictionary<int, DataPage<TItem>>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 항목 공급자 - ItemProvider


        /// <summary>
        /// 항목 공급자
        /// </summary>
        public IItemProvider<TItem> ItemProvider
        {
            get
            {
                return this.itemProvider;
            }
        }

        #endregion
        #region 페이지 크기 - PageSize

        /// <summary>
        /// 페이지 크기
        /// </summary>
        public int PageSize
        {
            get
            {
                return this.pageSize;
            }
        }

        #endregion
        #region 페이지 타임아웃 - PageTimeout

        /// <summary>
        /// 페이지 타임아웃
        /// </summary>
        public long PageTimeout
        {
            get
            {
                return this.pageTimeout;
            }
        }

        #endregion

        #region 카운트 - Count

        /// <summary>
        /// 카운트
        /// </summary>
        public int Count
        {
            get
            {
                if(this.count == -1)
                {
                    this.count = 0;

                    LoadCount();
                }

                return this.count;
            }
            protected set
            {
                this.count = value;
            }
        }

        #endregion
        #region 인덱서 - this[index]

        /// <summary>
        /// 인덱서
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>데이터 래퍼</returns>
        public DataWrapper<TItem> this[int index]
        {
            get
            {
                int pageIndex  = index / PageSize;
                int pageOffset = index % PageSize;

                RequestPage(pageIndex);

                if(pageOffset > PageSize / 2 && pageIndex < Count / PageSize)
                {
                    RequestPage(pageIndex + 1);
                }

                if(pageOffset < PageSize / 2 && pageIndex > 0)
                {
                    RequestPage(pageIndex - 1);
                }

                CleanUpPageDictionary();

                return this.pageDictionary[pageIndex].ItemList[pageOffset];
            }
            set
            {
                throw new NotSupportedException();
            }
        }

        #endregion
        #region 인덱서 - IList.this[index]

        /// <summary>
        /// 인덱서
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>객체</returns>
        object IList.this[int index]
        {
            get
            {
                return this[index];
            }
            set
            {
                throw new NotSupportedException();
            }
        }

        #endregion

        #region 동기화 루트 - SyncRoot

        /// <summary>
        /// 동기화 루트
        /// </summary>
        public object SyncRoot
        {
            get
            {
                return this;
            }
        }

        #endregion
        #region 동기화 여부 - IsSynchronized

        /// <summary>
        /// 동기화 여부
        /// </summary>
		public bool IsSynchronized
		{
			get
            {
                return false;
            }
		}

        #endregion
        #region 읽기 전용 여부 - IsReadOnly

        /// <summary>
        /// 읽기 전용 여부
        /// </summary>
        public bool IsReadOnly
        {
            get
            {
                return true;
            }
        }

        #endregion
        #region 고정 크기 여부 - IsFixedSize

        /// <summary>
        /// 고정 크기 여부
        /// </summary>
        public bool IsFixedSize
        {
            get
            {
                return false;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - VirtualizingCollection(itemProvider, pageSize, pageTimeout)

        /// <summary>
        ///  생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        /// <param name="pageSize">페이지 크기</param>
        /// <param name="pageTimeout">페이지 타임아웃</param>
        public VirtualizingCollection(IItemProvider<TItem> itemProvider, int pageSize, int pageTimeout)
        {
            this.itemProvider = itemProvider;
            this.pageSize     = pageSize;
            this.pageTimeout  = pageTimeout;
        }

        #endregion
        #region 생성자 - VirtualizingCollection(itemProvider, pageSize)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        /// <param name="pageSize">페이지 크기</param>
        public VirtualizingCollection(IItemProvider<TItem> itemProvider, int pageSize)
        {
            this.itemProvider = itemProvider;
            this.pageSize     = pageSize;
        }

        #endregion
        #region 생성자 - VirtualizingCollection(itemProvider)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        public VirtualizingCollection(IItemProvider<TItem> itemProvider)
        {
            this.itemProvider = itemProvider;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 열거자 구하기 - GetEnumerator()

        /// <summary>
        /// 열거자 구하기
        /// </summary>
        /// <returns>데이터 래퍼 열거자</returns>
        public IEnumerator<DataWrapper<TItem>> GetEnumerator()
        {
            for(int i = 0; i < Count; i++)
            {
                yield return this[i];
            }
        }

        #endregion
        #region 열거자 구하기 - IEnumerable.GetEnumerator()

        /// <summary>
        /// 열거자 구하기
        /// </summary>
        /// <returns>열거자</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion
        #region 추가하기 - Add(wrapper)

        /// <summary>
        /// 추가하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        public void Add(DataWrapper<TItem> wrapper)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 추가하기 - IList.Add(wrapper)

        /// <summary>
        /// 추가하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>인덱스</returns>
        int IList.Add(object wrapper)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 포함 여부 구하기 - Contains(wrapper)

        /// <summary>
        /// 포함 여부 구하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>포함 여부</returns>
        public bool Contains(DataWrapper<TItem> wrapper)
        {
            foreach(DataPage<TItem> page in this.pageDictionary.Values)
            {
                if(page.ItemList.Contains(wrapper))
                {
                    return true;
                }
            }

            return false;
        }

        #endregion
        #region 포함 여부 구하기 - IList.Contains(wrapper)

        /// <summary>
        /// 포함 여부 구하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>포함 여부</returns>
        bool IList.Contains(object wrapper)
        {
            return Contains((DataWrapper<TItem>)wrapper);
        }

        #endregion
        #region 지우기 - Clear()

        /// <summary>
        /// 지우기
        /// </summary>
        public void Clear()
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 인덱스 구하기 - IndexOf(wrapper)

        /// <summary>
        /// 인덱스 구하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>인덱스</returns>
        public int IndexOf(DataWrapper<TItem> wrapper)
        {
            foreach(KeyValuePair<int, DataPage<TItem>> keyValuePair in this.pageDictionary)
            {
                int indexWithinPage = keyValuePair.Value.ItemList.IndexOf(wrapper);

                if(indexWithinPage != -1)
                {
                    return PageSize * keyValuePair.Key + indexWithinPage;
                }
            }

            return -1;
        }

        #endregion
        #region 인덱스 구하기 - IList.IndexOf(wrapper)

        /// <summary>
        /// 인덱스 구하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>인덱스</returns>
        int IList.IndexOf(object wrapper)
        {
            return IndexOf((DataWrapper<TItem>)wrapper);
        }

        #endregion
        #region 삽입하기 - Insert(index, wrapper)

        /// <summary>
        /// 삽입하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <param name="wrapper">데이터 래퍼</param>
        public void Insert(int index, DataWrapper<TItem> wrapper)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 삽입하기 - IList.Insert(index, wrapper)

        /// <summary>
        /// 삽입하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <param name="wrapper">데이터 래퍼</param>
        void IList.Insert(int index, object wrapper)
        {
            Insert(index, (DataWrapper<TItem>)wrapper);
        }

        #endregion
        #region 제거하기 - RemoveAt(index)

        /// <summary>
        /// 제거하기
        /// </summary>
        /// <param name="index">인덱스</param>
        public void RemoveAt(int index)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 제거하기 - Remove(wrapper)

        /// <summary>
        /// 제거하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        /// <returns>처리 결과</returns>
        public bool Remove(DataWrapper<TItem> wrapper)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 제거하기 - IList.Remove(wrapper)

        /// <summary>
        /// 제거하기
        /// </summary>
        /// <param name="wrapper">데이터 래퍼</param>
        void IList.Remove(object wrapper)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 복사하기 - CopyTo(itemArray, arrayIndex)

        /// <summary>
        /// 복사하기
        /// </summary>
        /// <param name="itemArray">항목 배열</param>
        /// <param name="arrayIndex">배열 인덱스</param>
        public void CopyTo(DataWrapper<TItem>[] itemArray, int arrayIndex)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 복사하기 - ICollection.CopyTo(array, index)

        /// <summary>
        /// 복사하기
        /// </summary>
        /// <param name="array">배열</param>
        /// <param name="index">인덱스</param>
        void ICollection.CopyTo(Array array, int index)
        {
            throw new NotSupportedException();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 카운트 가져오기 - FetchCount()

        /// <summary>
        /// 카운트 가져오기
        /// </summary>
        /// <returns>카운트</returns>
        protected int FetchCount()
        {
            return ItemProvider.FetchCount();
        }

        #endregion
        #region 카운트 로드하기 - LoadCount()

        /// <summary>
        /// 카운트 로드하기
        /// </summary>
        protected virtual void LoadCount()
        {
            Count = FetchCount();
        }

        #endregion
        #region 페이지 만들기 - PopulatePage(pageIndex, itemList)

        /// <summary>
        /// 페이지 만들기
        /// </summary>
        /// <param name="pageIndex">페이지 인덱스</param>
        /// <param name="itemList">항목 리스트</param>
        protected virtual void PopulatePage(int pageIndex, IList<TItem> itemList)
        {
            DataPage<TItem> page;

            if(this.pageDictionary.TryGetValue(pageIndex, out page))
            {
                page.Populate(itemList);
            }
        }

        #endregion
        #region 페이지 가져오기 - FetchPage(pageIndex, pageLength, count)

        /// <summary>
        /// 페이지 가져오기
        /// </summary>
        /// <param name="pageIndex">페이지 인덱스</param>
        /// <param name="pageLength">페이지 길이</param>
        /// <param name="count">카운트</param>
        /// <returns>원본 항목 리스트</returns>
        protected IList<TItem> FetchPage(int pageIndex, int pageLength, out int count)
        {
            return ItemProvider.FetchRange(pageIndex * PageSize, pageLength, out count);
        }

        #endregion
        #region 페이지 로드하기 - LoadPage(pageIndex, pageLength)

        /// <summary>
        /// 페이지 로드하기
        /// </summary>
        /// <param name="pageIndex">페이지 인덱스</param>
        /// <param name="pageLength">페이지 길이</param>
        protected virtual void LoadPage(int pageIndex, int pageLength)
        {
            int count = 0;

            PopulatePage(pageIndex, FetchPage(pageIndex, pageLength, out count));

            Count = count;
        }

        #endregion
        #region 페이지 요청하기 - RequestPage(pageIndex)

        /// <summary>
        /// 페이지 요청하기
        /// </summary>
        /// <param name="pageIndex">페이지 인덱스</param>
        protected virtual void RequestPage(int pageIndex)
        {
            if(!this.pageDictionary.ContainsKey(pageIndex))
            {
                int pageLength = Math.Min(this.PageSize, this.Count - pageIndex * this.PageSize);

                DataPage<TItem> dataPage = new DataPage<TItem>(pageIndex * this.PageSize, pageLength);

                this.pageDictionary.Add(pageIndex, dataPage);

                LoadPage(pageIndex, pageLength);
            }
            else
            {
                this.pageDictionary[pageIndex].LastUsedTime = DateTime.Now;
            }
        }

        #endregion
        #region 캐시 비우기 - EmptyCache()

        /// <summary>
        /// 캐시 비우기
        /// </summary>
        protected void EmptyCache()
        {
            this.pageDictionary = new Dictionary<int, DataPage<TItem>>();
        }

        #endregion
        #region 페이지 딕셔너리 정리하기 - CleanUpPageDictionary()

        /// <summary>
        /// 페이지 딕셔너리 정리하기
        /// </summary>
        private void CleanUpPageDictionary()
        {
            int[] keyArray = this.pageDictionary.Keys.ToArray();

            foreach(int key in keyArray)
            {
                if(key != 0 && (DateTime.Now - this.pageDictionary[key].LastUsedTime).TotalMilliseconds > PageTimeout)
                {
                    bool removePage = true;

                    DataPage<TItem> page;

                    if(this.pageDictionary.TryGetValue(key, out page))
                    {
                        removePage = !page.InUse;
                    }

                    if(removePage)
                    {
                        this.pageDictionary.Remove(key);
                    }
                }
            }
        }

        #endregion
    }
}