﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 커스텀 정렬 설명 리스트
        /// </summary>
        private readonly List<CustomSortDescription> descriptionList;

        /// <summary>
        /// 고객 공급자
        /// </summary>
        private CustomerProvider customerProvider;

        /// <summary>
        /// 페이지 크기
        /// </summary>
        private int pageSize = 100;

        /// <summary>
        /// 페이지 타임아웃
        /// </summary>
        private int pageTimeout = 5000;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Consturctor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            PresentationTraceSources.DataBindingSource.Switch.Level = SourceLevels.Off;

            InitializeComponent();

            string defaultSortColumnName = "CustomerSince";

            DataGridColumn defaultSortColumn = dataGrid.Columns.Single(dgc => GetColumnSortMemberPath(dgc) == defaultSortColumnName);

            this.descriptionList = new List<CustomSortDescription>
            {
                new CustomSortDescription
                {
                    PropertyName = defaultSortColumnName,
                    Direction    = ListSortDirection.Descending,
                    Column       = defaultSortColumn
                }
            };


            RefreshData();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 일자 범위 선택기 일자 변경시 처리하기 - dateRangePicker_DateChanged(sender, e)

        /// <summary>
        /// 일자 범위 선택기 일자 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void dateRangePicker_DateChanged(object sender, EventArgs e)
        {
            RefreshData();
        }

        #endregion
        #region 데이터 그리드 정렬 처리하기 - dataGrid_Sorting(sender, e)

        /// <summary>
        /// 데이터 그리드 정렬 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void dataGrid_Sorting(object sender, DataGridSortingEventArgs e)
        {
            ApplySortColumn(e.Column);

            e.Handled = true;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 컬럼 정렬 멤버 경로 구하기 - GetColumnSortMemberPath(column)

        /// <summary>
        /// 컬럼 정렬 멤버 경로 구하기
        /// </summary>
        /// <param name="column">컬럼</param>
        /// <returns>컬럼 정렬 멤버 경로</returns>
        private string GetColumnSortMemberPath(DataGridColumn column)
        {
            string prefixToRemove = "Item.";
            string fullSortColumn = DataGridHelper.GetSortMemberPath(column);
            string sortColumn     = fullSortColumn.Substring(prefixToRemove.Length);

            return sortColumn;
        }

        #endregion
        #region 현재 정렬 문자열 구하기 - GetCurrentSortString()

        /// <summary>
        /// 현재 정렬 문자열 구하기
        /// </summary>
        /// <returns>현재 정렬 문자열</returns>
        private string GetCurrentSortString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            string separator = string.Empty;

            foreach(CustomSortDescription description in this.descriptionList)
            {
                stringBuilder.Append(separator);
                stringBuilder.Append(description.PropertyName);
 
                if(description.Direction == ListSortDirection.Descending)
                {
                    stringBuilder = stringBuilder.Append(" DESC");
                }

                separator = ", ";
            }

            return stringBuilder.ToString();
        }

        #endregion
        #region 컬럼 정렬 방향 업데이트하기 - UpdateColumnSortDirection()

        /// <summary>
        /// 컬럼 정렬 방향 업데이트하기
        /// </summary>
        private void UpdateColumnSortDirection()
        {
            foreach(CustomSortDescription description in this.descriptionList)
            {
                description.Column.SortDirection = description.Direction;
            }
        }

        #endregion
        #region 데이터 갱신하기 - RefreshData()

        /// <summary>
        /// 데이터 갱신하기
        /// </summary>
        private void RefreshData()
        {
            string sortString = GetCurrentSortString();

            this.customerProvider = new CustomerProvider
            (
                this.dateRangePicker.FromDate,
                this.dateRangePicker.ToDate,
                sortString
            );

            AsyncVirtualizingCollection<Customer> collection = new AsyncVirtualizingCollection<Customer>
            (
                this.customerProvider,
                this.pageSize,
                this.pageTimeout
            );

            DataContext = collection;

            UpdateColumnSortDirection();

            this.dataGrid.SelectedIndex = 0;
        }

        #endregion
        #region 정렬 컬럼 적용하기 - ApplySortColumn(column)

        /// <summary>
        /// 정렬 컬럼 적용하기
        /// </summary>
        /// <param name="column">컬럼</param>
        private void ApplySortColumn(DataGridColumn column)
        {
            string sortColumn = GetColumnSortMemberPath(column);

            CustomSortDescription existingDescription = this.descriptionList.SingleOrDefault(sd => sd.PropertyName == sortColumn);

            if(existingDescription == null)
            {
                existingDescription = new CustomSortDescription
                {
                    PropertyName = sortColumn,
                    Direction    = ListSortDirection.Ascending,
                    Column       = column
                };

                this.descriptionList.Add(existingDescription);
            }
            else
            {
                existingDescription.Direction = (existingDescription.Direction == ListSortDirection.Ascending) ? ListSortDirection.Descending : ListSortDirection.Ascending;
            }

            bool isShiftPressed = (Keyboard.Modifiers & ModifierKeys.Shift) != 0;

            if(!isShiftPressed)
            {
                for(int i = this.descriptionList.Count - 1; i >= 0; i--)
                {
                    CustomSortDescription description = this.descriptionList[i];

                    if(description.PropertyName != sortColumn)
                    {
                        this.descriptionList.RemoveAt(i);
                    }
                }
            }

            RefreshData();
        }

        #endregion
    }
}